@@ -1,12 +1,19 @@
 from ctypes.util import find_library
-from sys import maxsize
+import logging
+from sys import maxsize, platform
 from typing import List, Tuple
 from os.path import isfile
 import os.path
 from glob import glob
 from os import environ
-from sys import platform
+import numbers
 from cffi import FFI
+from mip.exceptions import (
+    ParameterNotAvailable,
+    SolutionNotAvailable,
+    ProgrammingError,
+    InterfacingError,
+)
 from mip import (
     Model,
     Column,
@@ -29,7 +36,10 @@ from mip import (
     SearchEmphasis,
     LP_Method,
 )
+from mip.lists import EmptyVarSol, EmptyRowSol
+import mip
 
+logger = logging.getLogger(__name__)
 
 try:
     found = False
@@ -38,21 +48,16 @@ try:
     if "GUROBI_HOME" in environ:
         if platform.lower().startswith("win"):
             libfile = glob(
-                os.path.join(
-                    os.environ["GUROBI_HOME"], "bin\\gurobi[0-9][0-9].dll"
-                )
+                os.path.join(os.environ["GUROBI_HOME"], "bin\\gurobi[0-9][0-9].dll")
             )
         else:
             libfile = glob(
-                os.path.join(
-                    os.environ["GUROBI_HOME"], "lib/libgurobi[0-9][0-9].*"
-                )
+                os.path.join(os.environ["GUROBI_HOME"], "lib/libgurobi[0-9][0-9].*")
             )
             if not libfile:
                 libfile = glob(
                     os.path.join(
-                        os.environ["GUROBI_HOME"],
-                        "lib/libgurobi.so.[0-9].[0-9].*",
+                        os.environ["GUROBI_HOME"], "lib/libgurobi.so.[0-9].[0-9].*",
                     )
                 )
 
@@ -68,16 +73,14 @@ try:
     if lib_path is None:
         for major_ver in reversed(range(6, 10)):
             for minor_ver in reversed(range(0, 11)):
-                lib_path = find_library(
-                    "gurobi{}{}".format(major_ver, minor_ver)
-                )
+                lib_path = find_library("gurobi{}{}".format(major_ver, minor_ver))
                 if lib_path is not None:
                     break
             if lib_path is not None:
                 break
 
     if lib_path is None:
-        raise Exception(
+        raise FileNotFoundError(
             """Gurobi not found. Plase check if the
         Gurobi dynamic loadable library is reachable or define
         the environment variable GUROBI_HOME indicating the gurobi
@@ -87,7 +90,7 @@ try:
     ffi = FFI()
 
     grblib = ffi.dlopen(lib_path)
-    print("gurobi version {}.{} found".format(major_ver, minor_ver))
+    logger.warning("gurobi version {}.{} found".format(major_ver, minor_ver))
 except Exception:
     raise ImportError
 
@@ -328,9 +331,7 @@ GRB_OPTIMAL = 2
 
 
 class SolverGurobi(Solver):
-    def __init__(
-        self, model: Model, name: str, sense: str, modelp: CData = ffi.NULL
-    ):
+    def __init__(self, model: Model, name: str, sense: str, modelp: CData = ffi.NULL):
         """modelp should be informed if a model should not be created,
         but only allow access to an existing one"""
         super().__init__(model, name, sense)
@@ -350,9 +351,8 @@ class SolverGurobi(Solver):
             # creating Gurobi environment
             st = GRBloadenv(self._env, "".encode("utf-8"))
             if st != 0:
-                raise Exception(
-                    "Gurobi environment could not be loaded,\
-    check your license."
+                raise InterfacingError(
+                    "Gurobi environment could not be loaded, check your license."
                 )
             self._env = self._env[0]
 
@@ -370,7 +370,7 @@ class SolverGurobi(Solver):
                 ffi.NULL,
             )
             if st != 0:
-                raise Exception("Could not create Gurobi model")
+                raise InterfacingError("Could not create Gurobi model")
             self._model = self._model[0]
 
             # setting objective sense
@@ -397,7 +397,18 @@ class SolverGurobi(Solver):
         self.__name_space = ffi.new("char[{}]".format(MAX_NAME_SIZE))
         self.__log = []
 
-        self.__x = None
+        # where solution will be stored
+        self.__x = EmptyVarSol(model)
+        self.__rc = EmptyVarSol(model)
+        self.__pi = EmptyRowSol(model)
+        self.__obj_val = None
+
+    def __clear_sol(self):
+        model = self.model
+        self.__x = EmptyVarSol(model)
+        self.__rc = EmptyVarSol(model)
+        self.__pi = EmptyRowSol(model)
+        self.__obj_val = None
 
     def __del__(self):
         # freeing Gurobi model and environment
@@ -430,18 +441,12 @@ class SolverGurobi(Solver):
         vtype = var_type.encode("utf-8")
 
         st = GRBaddvar(
-            self._model,
-            nz,
-            vind,
-            vval,
-            obj,
-            lb,
-            ub,
-            vtype,
-            name.encode("utf-8"),
+            self._model, nz, vind, vval, obj, lb, ub, vtype, name.encode("utf-8"),
         )
         if st != 0:
-            raise Exception("Error adding variable {} to model.".format(name))
+            raise ParameterNotAvailable(
+                "Error adding variable {} to model.".format(name)
+            )
 
         self.__n_cols_buffer += 1
         if vtype == BINARY or vtype == INTEGER:
@@ -465,11 +470,9 @@ class SolverGurobi(Solver):
         if not name:
             name = "r({})".format(self.num_rows())
 
-        st = GRBaddconstr(
-            self._model, nz, cind, cval, sense, rhs, name.encode("utf-8")
-        )
+        st = GRBaddconstr(self._model, nz, cind, cval, sense, rhs, name.encode("utf-8"))
         if st != 0:
-            raise Exception(
+            raise ParameterNotAvailable(
                 "Error adding constraint {} to the model".format(name)
             )
         self.__n_rows_buffer += 1
@@ -489,7 +492,7 @@ class SolverGurobi(Solver):
         w = ffi.new("double[]", [f for (v, f) in sos])
         st = GRBaddsos(self._model, 1, len(sos), types, beg, idx, w)
         if st != 0:
-            raise Exception("Error adding SOS to the model")
+            raise ParameterNotAvailable("Error adding SOS to the model")
 
     def get_objective_bound(self) -> float:
         return self.get_dbl_attr("ObjBound")
@@ -503,7 +506,7 @@ class SolverGurobi(Solver):
 
         st = GRBgetdblattrarray(self._model, attr, 0, self.num_cols(), obj)
         if st != 0:
-            raise Exception("Error getting objective function")
+            raise ParameterNotAvailable("Error getting objective function")
         obj_expr = xsum(
             obj[i] * self.model.vars[i]
             for i in range(self.num_cols())
@@ -517,11 +520,7 @@ class SolverGurobi(Solver):
 
     def relax(self):
         self.flush_cols()
-        idxv = [
-            var.idx
-            for var in self.model.vars
-            if var.var_type in [BINARY, INTEGER]
-        ]
+        idxv = [var.idx for var in self.model.vars if var.var_type in [BINARY, INTEGER]]
 
         n = len(idxv)
         idxs = ffi.new("int[]", idxv)
@@ -588,10 +587,8 @@ class SolverGurobi(Solver):
                                     difl = abs(obj_bound - log[-1][1][0])
                                     difu = abs(obj_best - log[-1][1][1])
                                     if difl >= 1e-6 or difu >= 1e-6:
-                                        print(
-                                            ">>>>>>> {} {}".format(
-                                                obj_bound, obj_best
-                                            )
+                                        logger.info(
+                                            ">>>>>>> {} {}".format(obj_bound, obj_best)
                                         )
                                     log.append((sec, (obj_bound, obj_best)))
 
@@ -602,7 +599,7 @@ class SolverGurobi(Solver):
 
                 error = GRBcbget(p_cbdata, where, GRB_CB_MIPNODE_STATUS, st)
                 if error:
-                    raise Exception("Could not get gurobi status")
+                    raise ParameterNotAvailable("Could not get gurobi status")
                 if st[0] != GRB_OPTIMAL:
                     return 0
 
@@ -662,83 +659,113 @@ class SolverGurobi(Solver):
         self.set_int_param("PoolSolutions", self.model.sol_pool_size)
 
         # executing Gurobi to solve the formulation
-        self.__x = None
+        self.__clear_sol()
+
+        # if solve only LP relax, saving var types
+        int_vars = []
+        if relax:
+            int_vars = [
+                (v, v.var_type)
+                for v in self.model.vars
+                if v.var_type in [BINARY, INTEGER]
+            ]
+            for v, _ in int_vars:
+                v.var_type = CONTINUOUS
+            self.update()
+
         status = GRBoptimize(self._model)
+        if int_vars:
+            for v, vt in int_vars:
+                v.var_type = vt
+            self.update()
+
         if status != 0:
             if status == 10009:
-                raise Exception(
-                    "gurobi found but license not accepted,\
-     please check it"
+                raise InterfacingError(
+                    "gurobi found but license not accepted, please check it"
                 )
             if status == 10001:
                 raise MemoryError("out of memory error")
 
-            raise Exception(
-                "Gurobi error {} while\
-                             optimizing.".format(
-                    status
-                )
-            )
+            raise InterfacingError("Gurobi error {} while optimizing.".format(status))
 
         status = self.get_int_attr("Status")
         # checking status for MIP optimization which
         # finished before the search to be
         # concluded (time, iteration limit...)
-        if self.num_int():
+        if (self.num_int() + self.get_int_attr("NumSOS")) and (not relax):
             if status in [8, 9, 10, 11, 13]:
                 nsols = self.get_int_attr("SolCount")
                 if nsols >= 1:
                     self.__x = ffi.new("double[{}]".format(self.num_cols()))
+                    self.__obj_val = self.get_dbl_attr("ObjVal")
+
                     attr = "X".encode("utf-8")
                     st = GRBgetdblattrarray(
                         self._model, attr, 0, self.num_cols(), self.__x
                     )
                     if st:
-                        raise Exception("Error quering Gurobi solution")
+                        raise ParameterNotAvailable("Error querying Gurobi solution")
 
                     return OptimizationStatus.FEASIBLE
-                else:
-                    return OptimizationStatus.NO_SOLUTION_FOUND
 
-        # todo: read solution status (code below is incomplete)
+                return OptimizationStatus.NO_SOLUTION_FOUND
+
         if status == 1:  # LOADED
             return OptimizationStatus.LOADED
         if status == 2:  # OPTIMAL
-            if self.__x is None:
+            if isinstance(self.__x, EmptyVarSol):
+                self.__obj_val = self.get_dbl_attr("ObjVal")
+
                 self.__x = ffi.new("double[{}]".format(self.num_cols()))
                 attr = "X".encode("utf-8")
-                st = GRBgetdblattrarray(
-                    self._model, attr, 0, self.num_cols(), self.__x
-                )
+                st = GRBgetdblattrarray(self._model, attr, 0, self.num_cols(), self.__x)
                 if st:
-                    raise Exception("Error quering Gurobi solution")
+                    raise ParameterNotAvailable("Error quering Gurobi solution")
+
+                if (self.num_int() + self.get_int_attr("NumSOS")) == 0 or (relax):
+                    self.__pi = ffi.new("double[{}]".format(self.num_rows()))
+                    attr = "Pi".encode("utf-8")
+                    st = GRBgetdblattrarray(
+                        self._model, attr, 0, self.num_rows(), self.__pi
+                    )
+                    if st:
+                        raise ParameterNotAvailable("Error quering Gurobi solution")
+
+                    self.__rc = ffi.new("double[{}]".format(self.num_cols()))
+                    attr = "RC".encode("utf-8")
+                    st = GRBgetdblattrarray(
+                        self._model, attr, 0, self.num_cols(), self.__rc
+                    )
+                    if st:
+                        raise ParameterNotAvailable("Error quering Gurobi solution")
 
             return OptimizationStatus.OPTIMAL
-        elif status == 3:  # INFEASIBLE
+        if status == 3:  # INFEASIBLE
             return OptimizationStatus.INFEASIBLE
-        elif status == 4:  # INF_OR_UNBD
+        if status == 4:  # INF_OR_UNBD
             return OptimizationStatus.UNBOUNDED
-        elif status == 5:  # UNBOUNDED
+        if status == 5:  # UNBOUNDED
             return OptimizationStatus.UNBOUNDED
-        elif status == 6:  # CUTOFF
+        if status == 6:  # CUTOFF
             return OptimizationStatus.CUTOFF
-        elif status == 7:  # ITERATION_LIMIT
+        if status == 7:  # ITERATION_LIMIT
             return OptimizationStatus.OTHER
-        elif status == 8:  # NODE_LIMIT
+        if status == 8:  # NODE_LIMIT
             return OptimizationStatus.OTHER
-        elif status == 9:  # TIME_LIMIT
+        if status == 9:  # TIME_LIMIT
             return OptimizationStatus.OTHER
-        elif status == 10:  # SOLUTION_LIMIT
+        if status == 10:  # SOLUTION_LIMIT
             return OptimizationStatus.FEASIBLE
-        elif status == 11:  # INTERRUPTED
+        if status == 11:  # INTERRUPTED
             return OptimizationStatus.OTHER
-        elif status == 12:  # NUMERIC
+        if status == 12:  # NUMERIC
             return OptimizationStatus.OTHER
-        elif status == 13:  # SUBOPTIMAL
+        if status == 13:  # SUBOPTIMAL
             return OptimizationStatus.FEASIBLE
-        elif status == 14:  # INPROGRESS
+        if status == 14:  # INPROGRESS
             return OptimizationStatus.OTHER
-        elif status == 15:  # USER_OBJ_LIMIT
+        if status == 15:  # USER_OBJ_LIMIT
             return OptimizationStatus.FEASIBLE
 
         self._updated = True
@@ -751,7 +778,7 @@ class SolverGurobi(Solver):
         elif isense == -1:
             return MAXIMIZE
         else:
-            raise Exception("Unknown sense")
+            raise ValueError("Unknown sense")
 
     def set_objective_sense(self, sense: str):
         if sense.strip().upper() == MAXIMIZE.strip().upper():
@@ -759,10 +786,8 @@ class SolverGurobi(Solver):
         elif sense.strip().upper() == MINIMIZE.strip().upper():
             self.set_int_attr("ModelSense", 1)
         else:
-            raise Exception(
-                "Unknown sense: {}, use {} or {}".format(
-                    sense, MAXIMIZE, MINIMIZE
-                )
+            raise ValueError(
+                "Unknown sense: {}, use {} or {}".format(sense, MAXIMIZE, MINIMIZE)
             )
         self.__updated = False
 
@@ -778,7 +803,7 @@ class SolverGurobi(Solver):
         idx = ffi.new("int *")
         st = GRBgetvarbyname(self._model, name.encode("utf-8"), idx)
         if st:
-            raise Exception("Error calling GRBgetvarbyname")
+            raise ParameterNotAvailable("Error calling GRBgetvarbyname")
         return idx[0]
 
     def get_objective_value_i(self, i: int) -> float:
@@ -786,13 +811,18 @@ class SolverGurobi(Solver):
         return self.get_dbl_attr("PoolObjVal")
 
     def get_objective_value(self) -> float:
-        return self.get_dbl_attr("ObjVal")
+        return self.__obj_val
 
     def get_log(self) -> List[Tuple[float, Tuple[float, float]]]:
         return self.__log
 
     def set_processing_limits(
-        self, max_time: float = INF, max_nodes: float = INF, max_sol: int = INF
+        self: "Solver",
+        max_time: numbers.Real = mip.INF,
+        max_nodes: int = mip.INT_MAX,
+        max_sol: int = mip.INT_MAX,
+        max_seconds_same_incumbent: float = mip.INF,
+        max_nodes_same_incumbent: int = mip.INT_MAX,
     ):
         # todo: Set limits even when they are 'inf'
         if max_time != INF:
@@ -820,7 +850,9 @@ class SolverGurobi(Solver):
         st = GRBsetdblattrarray(self._model, attr, 0, num_vars, zeros)
 
         if st != 0:
-            raise Exception("Could not set gurobi double attribute array Obj")
+            raise ParameterNotAvailable(
+                "Could not set gurobi double attribute array Obj"
+            )
 
         # setting objective sense
         if MAXIMIZE in (lin_expr.sense, sense):
@@ -832,7 +864,7 @@ class SolverGurobi(Solver):
         self.set_dbl_attr("ObjCon", const)
         error = GRBsetdblattrlist(self._model, attr, nz, cind, cval)
         if error != 0:
-            raise Exception("Error modifying attribute Obj")
+            raise ParameterNotAvailable("Error modifying attribute Obj")
         self.__n_modified_cols += 1
 
     def set_objective_const(self, const: float) -> None:
@@ -845,11 +877,9 @@ class SolverGurobi(Solver):
         cind = ffi.new("int[]", [el[0].idx for el in start])
         cval = ffi.new("double[]", [el[1] for el in start])
 
-        st = GRBsetdblattrlist(
-            self._model, "Start".encode("utf-8"), nz, cind, cval
-        )
+        st = GRBsetdblattrlist(self._model, "Start".encode("utf-8"), nz, cind, cval)
         if st != 0:
-            raise Exception("Error modifying attribute Start")
+            raise ParameterNotAvailable("Error modifying attribute Start")
         self.__updated = False
 
     def flush_cols(self):
@@ -869,16 +899,16 @@ class SolverGurobi(Solver):
         self.update()
         st = GRBwrite(self._model, file_path.encode("utf-8"))
         if st != 0:
-            raise Exception("Could not write gurobi model.")
+            raise InterfacingError("Could not write gurobi model.")
 
     def read(self, file_path: str) -> None:
         if not isfile(file_path):
-            raise Exception("File {} does not exists".format(file_path))
+            raise FileNotFoundError("File {} does not exist".format(file_path))
         GRBfreemodel(self._model)
         self._model = ffi.new("GRBmodel **")
         st = GRBreadmodel(self._env, file_path.encode("utf-8"), self._model)
         if st != 0:
-            raise Exception(
+            raise InterfacingError(
                 "Could not read model {}, check contents".format(file_path)
             )
         self._model = self._model[0]
@@ -925,11 +955,9 @@ class SolverGurobi(Solver):
 
         nnz = ffi.new("int *")
         # obtaining number of non-zeros
-        st = GRBgetconstrs(
-            self._model, nnz, ffi.NULL, ffi.NULL, ffi.NULL, constr.idx, 1
-        )
+        st = GRBgetconstrs(self._model, nnz, ffi.NULL, ffi.NULL, ffi.NULL, constr.idx, 1)
         if st != 0:
-            raise Exception(
+            raise ParameterNotAvailable(
                 "Could not get info for constraint {}".format(constr.idx)
             )
         nz = nnz[0]
@@ -942,7 +970,7 @@ class SolverGurobi(Solver):
         # obtaining variables and coefficients
         st = GRBgetconstrs(self._model, nnz, cbeg, cind, cval, constr.idx, 1)
         if st != 0:
-            raise Exception("Could not query constraint contents")
+            raise ParameterNotAvailable("Could not query constraint contents")
 
         # obtaining sense and rhs
         c_sense = ffi.new("char *")
@@ -951,14 +979,12 @@ class SolverGurobi(Solver):
             self._model, "Sense".encode("utf-8"), constr.idx, c_sense
         )
         if st != 0:
-            raise Exception(
+            raise ParameterNotAvailable(
                 "Could not query sense for constraint {}".format(constr.idx)
             )
-        st = GRBgetdblattrelement(
-            self._model, "RHS".encode("utf-8"), constr.idx, rhs
-        )
+        st = GRBgetdblattrelement(self._model, "RHS".encode("utf-8"), constr.idx, rhs)
         if st != 0:
-            raise Exception(
+            raise ParameterNotAvailable(
                 "Could not query RHS for constraint {}".format(constr.idx)
             )
 
@@ -996,21 +1022,21 @@ class SolverGurobi(Solver):
         return self.get_dbl_attr_element("Slack", constr.idx)
 
     def constr_get_pi(self, constr: "Constr") -> float:
-        return self.get_dbl_attr_element("Pi", constr.idx)
+        return self.__pi[constr.idx]
 
     def constr_get_index(self, name: str) -> int:
         GRBupdatemodel(self._model)
         idx = ffi.new("int *")
         st = GRBgetconstrbyname(self._model, name.encode("utf-8"), idx)
         if st != 0:
-            raise Exception("Error calling GRBgetconstrbyname")
+            raise ParameterNotAvailable("Error calling GRBgetconstrbyname")
         return idx[0]
 
     def remove_constrs(self, constrsList: List[int]):
         idx = ffi.new("int[]", constrsList)
         st = GRBdelconstrs(self._model, len(constrsList), idx)
         if st != 0:
-            raise Exception("Error calling GRBdelconstrs")
+            raise ParameterNotAvailable("Error calling GRBdelconstrs")
         self.__n_modified_rows += len(constrsList)
 
     def var_get_lb(self, var: Var) -> float:
@@ -1040,11 +1066,9 @@ class SolverGurobi(Solver):
     def var_get_var_type(self, var: Var) -> str:
         self.flush_cols()
         res = ffi.new("char *")
-        st = GRBgetcharattrelement(
-            self._model, "VType".encode("utf-8"), var.idx, res
-        )
+        st = GRBgetcharattrelement(self._model, "VType".encode("utf-8"), var.idx, res)
         if st != 0:
-            raise Exception("Error querying variable type in gurobi")
+            raise ParameterNotAvailable("Error querying variable type in gurobi")
 
         vt = res[0].decode("utf-8")
 
@@ -1067,11 +1091,9 @@ class SolverGurobi(Solver):
         nnz = ffi.new("int*")
 
         # obtaining number of non-zeros
-        error = GRBgetvars(
-            self._model, nnz, ffi.NULL, ffi.NULL, ffi.NULL, var.idx, 1
-        )
+        error = GRBgetvars(self._model, nnz, ffi.NULL, ffi.NULL, ffi.NULL, var.idx, 1)
         if error != 0:
-            raise Exception("Error querying gurobi model information")
+            raise ParameterNotAvailable("Error querying gurobi model information")
 
         nz = nnz[0]
 
@@ -1083,7 +1105,7 @@ class SolverGurobi(Solver):
         # obtaining variables and coefficients
         error = GRBgetvars(self._model, nnz, cbeg, cind, cval, var.idx, 1)
         if error != 0:
-            raise Exception("Error querying gurobi model information")
+            raise ParameterNotAvailable("Error querying gurobi model information")
 
         constr = [self.model.constrs[cind[i]] for i in range(nz)]
         coefs = [float(cval[i]) for i in range(nz)]
@@ -1094,7 +1116,7 @@ class SolverGurobi(Solver):
         raise NotImplementedError("Gurobi functionality currently unavailable")
 
     def var_get_rc(self, var: Var) -> float:
-        return self.get_dbl_attr_element("RC", var.idx)
+        return self.__rc[var.idx]
 
     def var_get_x(self, var: Var) -> float:
         return self.__x[var.idx]
@@ -1107,7 +1129,7 @@ class SolverGurobi(Solver):
         idx = ffi.new("int[]", varsList)
         st = GRBdelvars(self._model, len(varsList), idx)
         if st != 0:
-            raise Exception("Error calling GRBdelvars")
+            raise ParameterNotAvailable("Error calling GRBdelvars")
         self.__n_modified_cols += len(varsList)
 
     def get_emphasis(self) -> SearchEmphasis:
@@ -1145,12 +1167,13 @@ class SolverGurobi(Solver):
         self.__updated = True
 
     def set_char_attr_element(self, name: str, index: int, value: str):
-        assert len(value) == 1
+        if len(value) != 1:
+            raise ValueError("Expected value of length 1, got {}".format(len(value)))
         error = GRBsetcharattrelement(
             self._model, name.encode("utf-8"), index, value.encode("utf-8")
         )
         if error != 0:
-            raise Exception(
+            raise ParameterNotAvailable(
                 "Error setting gurobi char attr element {} index {} to value {}".format(
                     name, index, value
                 )
@@ -1158,34 +1181,26 @@ class SolverGurobi(Solver):
 
     def get_dbl_attr_element(self, name: str, index: int) -> float:
         res = ffi.new("double *")
-        error = GRBgetdblattrelement(
-            self._model, name.encode("utf-8"), index, res
-        )
+        error = GRBgetdblattrelement(self._model, name.encode("utf-8"), index, res)
         if error != 0:
-            raise Exception(
-                "Error get grb double attr element {} index {}".format(
-                    name, index
-                )
+            raise ParameterNotAvailable(
+                "Error get grb double attr element {} index {}".format(name, index)
             )
         return res[0]
 
     def set_dbl_attr_element(self, name: str, index: int, value: float):
-        error = GRBsetdblattrelement(
-            self._model, name.encode("utf-8"), index, value
-        )
+        error = GRBsetdblattrelement(self._model, name.encode("utf-8"), index, value)
         if error != 0:
-            raise Exception(
+            raise ParameterNotAvailable(
                 "Error modifying dbl attribute {} for element {} to value {}".format(
                     name, index, value
                 )
             )
 
     def set_int_attr_element(self, name: str, index: int, value: int):
-        error = GRBsetintattrelement(
-            self._model, name.encode("utf-8"), index, value
-        )
+        error = GRBsetintattrelement(self._model, name.encode("utf-8"), index, value)
         if error != 0:
-            raise Exception(
+            raise ParameterNotAvailable(
                 "Error modifying int attribute {} for element {} to value {}".format(
                     name, index, value
                 )
@@ -1194,7 +1209,7 @@ class SolverGurobi(Solver):
     def set_int_attr(self, name: str, value: int):
         error = GRBsetintattr(self._model, name.encode("utf-8"), value)
         if error != 0:
-            raise Exception(
+            raise ParameterNotAvailable(
                 "Error modifying int attribute {} to {}".format(name, value)
             )
         GRBupdatemodel(self._model)
@@ -1202,7 +1217,7 @@ class SolverGurobi(Solver):
     def set_dbl_attr(self, name: str, value: float):
         error = GRBsetdblattr(self._model, name.encode("utf-8"), value)
         if error != 0:
-            raise Exception(
+            raise ParameterNotAvailable(
                 "Error modifying double attribute {} to {}".format(name, value)
             )
 
@@ -1210,7 +1225,7 @@ class SolverGurobi(Solver):
         res = ffi.new("int *")
         error = GRBgetintattr(self._model, name.encode("utf-8"), res)
         if error != 0:
-            raise Exception("Error getting int attribute {}".format(name))
+            raise ParameterNotAvailable("Error getting int attribute {}".format(name))
         return res[0]
 
     def get_int_param(self, name: str) -> int:
@@ -1218,7 +1233,7 @@ class SolverGurobi(Solver):
         env = GRBgetenv(self._model)
         error = GRBgetintparam(env, name.encode("utf-8"), res)
         if error != 0:
-            raise Exception(
+            raise ParameterNotAvailable(
                 "Error getting gurobi integer parameter {}".format(name)
             )
         return res[0]
@@ -1227,10 +1242,8 @@ class SolverGurobi(Solver):
         env = GRBgetenv(self._model)
         error = GRBsetintparam(env, name.encode("utf-8"), value)
         if error != 0:
-            raise Exception(
-                "Error mofifying int parameter {} to value {}".format(
-                    name, value
-                )
+            raise ParameterNotAvailable(
+                "Error mofifying int parameter {} to value {}".format(name, value)
             )
         GRBupdatemodel(self._model)
 
@@ -1238,7 +1251,7 @@ class SolverGurobi(Solver):
         res = ffi.new("double *")
         error = GRBgetdblattr(self._model, attr.encode("utf-8"), res)
         if error != 0:
-            raise Exception(
+            raise ParameterNotAvailable(
                 "Error getting gurobi double attribute {}".format(attr)
             )
         return res[0]
@@ -1247,10 +1260,8 @@ class SolverGurobi(Solver):
         env = GRBgetenv(self._model)
         error = GRBsetdblparam(env, param.encode("utf-8"), float(value))
         if error != 0:
-            raise Exception(
-                "Error setting gurobi double param "
-                + param
-                + " to {}".format(value)
+            raise ParameterNotAvailable(
+                "Error setting gurobi double param {}  to {}".format(param, value)
             )
 
     def get_dbl_param(self, param: str) -> float:
@@ -1258,18 +1269,16 @@ class SolverGurobi(Solver):
         env = GRBgetenv(self._model)
         error = GRBgetdblparam(env, param.encode("utf-8"), res)
         if error != 0:
-            raise Exception(
+            raise ParameterNotAvailable(
                 "Error getting gurobi double parameter {}".format(param)
             )
         return res[0]
 
     def get_str_attr_element(self, attr: str, index: int) -> str:
         vName = ffi.new("char **")
-        error = GRBgetstrattrelement(
-            self._model, attr.encode("utf-8"), index, vName
-        )
+        error = GRBgetstrattrelement(self._model, attr.encode("utf-8"), index, vName)
         if error != 0:
-            raise Exception(
+            raise ParameterNotAvailable(
                 "Error getting str attribute {} index {}".format(attr, index)
             )
         return ffi.string(vName[0]).decode("utf-8")
@@ -1278,7 +1287,7 @@ class SolverGurobi(Solver):
         vName = ffi.new("char **")
         error = GRBgetstrattr(self._model, "ModelName".encode("utf-8"), vName)
         if error != 0:
-            raise Exception("Error getting problem name from gurobi")
+            raise ParameterNotAvailable("Error getting problem name from gurobi")
         return ffi.string(vName[0]).decode("utf-8")
 
     def set_problem_name(self, name: str):
@@ -1287,7 +1296,7 @@ class SolverGurobi(Solver):
         )
 
         if error != 0:
-            raise Exception("Error setting problem name in Gurobi")
+            raise ParameterNotAvailable("Error setting problem name in Gurobi")
         GRBupdatemodel(self._model)
 
     def get_pump_passes(self) -> int:
@@ -1309,6 +1318,7 @@ class SolverGurobiCB(SolverGurobi):
         cb_data: CData = ffi.NULL,
         where: int = -1,
     ):
+        # TODO: Replace with exceptions with meaningful text.
         assert grb_model != ffi.NULL
         assert cb_data != ffi.NULL
 
@@ -1332,20 +1342,18 @@ class SolverGurobiCB(SolverGurobi):
         if where == GRB_CB_MIPNODE:
             res = GRBcbget(cb_data, where, GRB_CB_MIPNODE_STATUS, gstatus)
             if res != 0:
-                raise Exception("Error getting status")
+                raise ParameterNotAvailable("Error getting status")
             if gstatus[0] == GRB_OPTIMAL:
                 self._status = OptimizationStatus.OPTIMAL
                 ires = ffi.new("int *")
                 st = GRBgetintattr(grb_model, "NumVars".encode("utf-8"), ires)
                 if st != 0:
-                    raise Exception("Could not query number of variables")
+                    raise ParameterNotAvailable("Could not query number of variables")
                 ncols = ires[0]
                 self._cb_sol = ffi.new("double[{}]".format(ncols))
-                res = GRBcbget(
-                    cb_data, where, GRB_CB_MIPNODE_REL, self._cb_sol
-                )
+                res = GRBcbget(cb_data, where, GRB_CB_MIPNODE_REL, self._cb_sol)
                 if res != 0:
-                    raise Exception("Error getting fractional solution")
+                    raise ParameterNotAvailable("Error getting fractional solution")
             else:
                 self._cb_sol = ffi.NULL
         elif where == GRB_CB_MIPSOL:
@@ -1353,7 +1361,7 @@ class SolverGurobiCB(SolverGurobi):
             ires = ffi.new("int *")
             st = GRBgetintattr(grb_model, "NumVars".encode("utf-8"), ires)
             if st != 0:
-                raise Exception(
+                raise ParameterNotAvailable(
                     "Could not query number of variables in Gurobi callback"
                 )
             ncols = ires[0]
@@ -1361,14 +1369,13 @@ class SolverGurobiCB(SolverGurobi):
             self._cb_sol = ffi.new("double[{}]".format(ncols))
             res = GRBcbget(cb_data, where, GRB_CB_MIPSOL_SOL, self._cb_sol)
             if res != 0:
-                raise Exception(
-                    "Error getting integer solution in gurobi \
-                                callback"
+                raise ParameterNotAvailable(
+                    "Error getting integer solution in gurobi callback"
                 )
             objp = ffi.new("double *")
             res = GRBcbget(cb_data, where, GRB_CB_MIPSOL_OBJ, objp)
             if res != 0:
-                raise Exception("Error getting solution obj in Gurobi")
+                raise ParameterNotAvailable("Error getting solution obj in Gurobi")
             self._obj_value = objp[0]
 
         else:
@@ -1402,20 +1409,18 @@ class SolverGurobiCB(SolverGurobi):
         if self._where == GRB_CB_MIPSOL:
             res = GRBcblazy(self._cb_data, nz, cind, cval, sense, rhs)
             if res != 0:
-                raise Exception("Error adding lazy constraint in Gurobi.")
+                raise ParameterNotAvailable("Error adding lazy constraint in Gurobi.")
         elif self._where == GRB_CB_MIPNODE:
             res = GRBcbcut(self._cb_data, nz, cind, cval, sense, rhs)
             if res != 0:
-                raise Exception("Error adding lazy constraint in Gurobi.")
+                raise ParameterNotAvailable("Error adding lazy constraint in Gurobi.")
 
     def add_lazy_constr(self, lin_expr: "LinExpr"):
         if self._where == GRB_CB_MIPSOL:
             # collecting linear expression data
             nz = len(lin_expr.expr)
             cind = ffi.new("int[]", [var.idx for var in lin_expr.expr.keys()])
-            cval = ffi.new(
-                "double[]", [coef for coef in lin_expr.expr.values()]
-            )
+            cval = ffi.new("double[]", [coef for coef in lin_expr.expr.values()])
 
             # constraint sense and rhs
             sense = lin_expr.sense.encode("utf-8")
@@ -1423,14 +1428,12 @@ class SolverGurobiCB(SolverGurobi):
 
             res = GRBcblazy(self._cb_data, nz, cind, cval, sense, rhs)
             if res != 0:
-                raise Exception("Error adding lazy constraint in Gurobi.")
+                raise ParameterNotAvailable("Error adding lazy constraint in Gurobi.")
         elif self._where == GRB_CB_MIPNODE:
             # collecting linear expression data
             nz = len(lin_expr.expr)
             cind = ffi.new("int[]", [var.idx for var in lin_expr.expr.keys()])
-            cval = ffi.new(
-                "double[]", [coef for coef in lin_expr.expr.values()]
-            )
+            cval = ffi.new("double[]", [coef for coef in lin_expr.expr.values()])
 
             # constraint sense and rhs
             sense = lin_expr.sense.encode("utf-8")
@@ -1438,11 +1441,10 @@ class SolverGurobiCB(SolverGurobi):
 
             res = GRBcbcut(self._cb_data, nz, cind, cval, sense, rhs)
             if res != 0:
-                raise Exception("Error adding cutting plane in Gurobi.")
+                raise ParameterNotAvailable("Error adding cutting plane in Gurobi.")
         else:
-            raise Exception(
-                "Calling add_lazy_constr in unknow callback \
-                            location"
+            raise ProgrammingError(
+                "Calling add_lazy_constr in unknown callback location"
             )
 
     def get_status(self):
@@ -1450,7 +1452,7 @@ class SolverGurobiCB(SolverGurobi):
 
     def var_get_x(self, var: Var):
         if self._cb_sol == ffi.NULL:
-            raise Exception("Solution not available")
+            raise SolutionNotAvailable("Solution not available")
 
         return self._cb_sol[var.idx]
 
@@ -1460,10 +1462,7 @@ class SolverGurobiCB(SolverGurobi):
 
 class ModelGurobiCB(Model):
     def __init__(
-        self,
-        grb_model: CData = ffi.NULL,
-        cb_data: CData = ffi.NULL,
-        where: int = -1,
+        self, grb_model: CData = ffi.NULL, cb_data: CData = ffi.NULL, where: int = -1,
     ):
         # initializing variables with default values
         self.solver_name = "gurobicb"
